# Copyright 2016 Red Hat.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import os
import resource
import shutil
import sys

USAGE_PROGRAM = (
    f'{os.path.basename(sys.executable)} -m oslo_concurrency.prlimit'
)

RESOURCES = (
    # argparse argument => resource
    ('as', resource.RLIMIT_AS),
    ('core', resource.RLIMIT_CORE),
    ('cpu', resource.RLIMIT_CPU),
    ('data', resource.RLIMIT_DATA),
    ('fsize', resource.RLIMIT_FSIZE),
    ('memlock', resource.RLIMIT_MEMLOCK),
    ('nofile', resource.RLIMIT_NOFILE),
    ('nproc', resource.RLIMIT_NPROC),
    ('rss', resource.RLIMIT_RSS),
    ('stack', resource.RLIMIT_STACK),
)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='prlimit', prog=USAGE_PROGRAM)
    parser.add_argument('--as', type=int, help='Address space limit in bytes')
    parser.add_argument(
        '--core', type=int, help='Core file size limit in bytes'
    )
    parser.add_argument('--cpu', type=int, help='CPU time limit in seconds')
    parser.add_argument('--data', type=int, help='Data size limit in bytes')
    parser.add_argument('--fsize', type=int, help='File size limit in bytes')
    parser.add_argument(
        '--memlock', type=int, help='Locked memory limit in bytes'
    )
    parser.add_argument(
        '--nofile', type=int, help='Maximum number of open files'
    )
    parser.add_argument(
        '--nproc', type=int, help='Maximum number of processes'
    )
    parser.add_argument(
        '--rss', type=int, help='Maximum Resident Set Size (RSS) in bytes'
    )
    parser.add_argument('--stack', type=int, help='Stack size limit in bytes')
    parser.add_argument('program', help='Program (absolute path)')
    parser.add_argument(
        'program_args', metavar="arg", nargs='...', help='Program parameters'
    )

    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()

    program = args.program
    if not os.path.isabs(program):
        # program uses a relative path: try to find the absolute path
        # to the executable
        program_abs = shutil.which(program)
        if program_abs:
            program = program_abs

    for arg_name, rlimit in RESOURCES:
        value = getattr(args, arg_name)
        if value is None:
            continue
        try:
            resource.setrlimit(rlimit, (value, value))
        except ValueError as exc:
            print(
                f"{USAGE_PROGRAM}: failed to set the {arg_name.upper()} "
                f"resource limit: {exc}",
                file=sys.stderr,
            )
            sys.exit(1)

    try:
        # it's okay to start this user-provided program
        os.execv(program, [program] + args.program_args)  # noqa: S606
    except Exception as exc:
        print(
            f"{USAGE_PROGRAM}: failed to execute {program}: {exc}",
            file=sys.stderr,
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
System-level utilities and helper functions.
"""

from collections.abc import Callable, Iterable
import functools
import logging
import multiprocessing
import os
import random
import shlex
import signal
import subprocess
import sys
import time
from typing import Any, Literal, overload, TYPE_CHECKING

import enum
from oslo_utils import encodeutils
from oslo_utils import strutils
from oslo_utils import timeutils

from oslo_concurrency._i18n import _

if TYPE_CHECKING:
    import paramiko.client


LOG = logging.getLogger(__name__)


class InvalidArgumentError(Exception):
    def __init__(self, message: object = None) -> None:
        super().__init__(message)


class UnknownArgumentError(Exception):
    def __init__(self, message: object = None) -> None:
        super().__init__(message)


class ProcessExecutionError(Exception):
    def __init__(
        self,
        stdout: str | None = None,
        stderr: str | None = None,
        exit_code: str | int | None = None,
        cmd: list[str] | str | None = None,
        description: str | None = None,
    ):
        super().__init__(stdout, stderr, exit_code, cmd, description)
        self.exit_code = exit_code
        self.stderr = stderr
        self.stdout = stdout
        self.cmd = cmd
        self.description = description

    def __str__(self) -> str:
        description = self.description
        if description is None:
            description = _("Unexpected error while running command.")

        exit_code = self.exit_code
        if exit_code is None:
            exit_code = '-'

        message = _(
            '%(description)s\n'
            'Command: %(cmd)s\n'
            'Exit code: %(exit_code)s\n'
            'Stdout: %(stdout)r\n'
            'Stderr: %(stderr)r'
        ) % {
            'description': description,
            'cmd': self.cmd,
            'exit_code': exit_code,
            'stdout': self.stdout,
            'stderr': self.stderr,
        }
        return message


class NoRootWrapSpecified(Exception):
    def __init__(self, message: object = None) -> None:
        super().__init__(message)


def _subprocess_setup(on_preexec_fn: Callable[..., Any] | None) -> None:
    # Python installs a SIGPIPE handler by default. This is usually not what
    # non-Python subprocesses expect.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    if on_preexec_fn:
        on_preexec_fn()


@enum.unique
class LogErrors(enum.IntEnum):
    """Enumerations that affect if stdout and stderr are logged on error.

    .. versionadded:: 2.7
    """

    #: No logging on errors.
    DEFAULT = 0

    #: Log an error on **each** occurence of an error.
    ALL = 1

    #: Log an error on the last attempt that errored **only**.
    FINAL = 2


# Retain these aliases for a number of releases...
LOG_ALL_ERRORS = LogErrors.ALL
LOG_FINAL_ERROR = LogErrors.FINAL
LOG_DEFAULT_ERROR = LogErrors.DEFAULT


class ProcessLimits:
    """Resource limits on a process.

    Attributes:

    * address_space: Address space limit in bytes
    * core_file_size: Core file size limit in bytes
    * cpu_time: CPU time limit in seconds
    * data_size: Data size limit in bytes
    * file_size: File size limit in bytes
    * memory_locked: Locked memory limit in bytes
    * number_files: Maximum number of open files
    * number_processes: Maximum number of processes
    * resident_set_size: Maximum Resident Set Size (RSS) in bytes
    * stack_size: Stack size limit in bytes

    This object can be used for the *prlimit* parameter of :func:`execute`.
    """

    _LIMITS = {
        "address_space": "--as",
        "core_file_size": "--core",
        "cpu_time": "--cpu",
        "data_size": "--data",
        "file_size": "--fsize",
        "memory_locked": "--memlock",
        "number_files": "--nofile",
        "number_processes": "--nproc",
        "resident_set_size": "--rss",
        "stack_size": "--stack",
    }

    def __init__(
        self,
        *,
        address_space: int | None = None,
        core_file_size: int | None = None,
        cpu_time: int | None = None,
        data_size: int | None = None,
        file_size: int | None = None,
        memory_locked: int | None = None,
        number_files: int | None = None,
        number_processes: int | None = None,
        resident_set_size: int | None = None,
        stack_size: int | None = None,
    ):
        self.address_space = address_space
        self.core_file_size = core_file_size
        self.cpu_time = cpu_time
        self.data_size = data_size
        self.file_size = file_size
        self.memory_locked = memory_locked
        self.number_files = number_files
        self.number_processes = number_processes
        self.resident_set_size = resident_set_size
        self.stack_size = stack_size

    def prlimit_args(self) -> list[str]:
        """Create a list of arguments for the prlimit command line."""
        args = []
        for limit in self._LIMITS:
            val = getattr(self, limit)
            if val is not None:
                args.append(f"{self._LIMITS[limit]}={val}")
        return args


@overload
def execute(
    *cmd: str,
    cwd: str | None = None,
    process_input: str | bytes | None = None,
    env_variables: dict[str, str] | None = None,
    check_exit_code: bool | int | Iterable[int] | None = 0,
    delay_on_retry: bool = True,
    attempts: int = 1,
    run_as_root: bool = False,
    root_helper: str = '',
    shell: bool = False,
    loglevel: int = logging.DEBUG,
    log_errors: LogErrors = LogErrors.DEFAULT,
    binary: Literal[False] = ...,
    on_execute: Callable[..., Any] | None = None,
    on_completion: Callable[..., Any] | None = None,
    preexec_fn: Callable[..., Any] | None = None,
    prlimit: ProcessLimits | None = None,
    python_exec: str | None = None,
    timeout: int | None = None,
) -> tuple[str, str]: ...


@overload
def execute(
    *cmd: str,
    cwd: str | None = None,
    process_input: str | bytes | None = None,
    env_variables: dict[str, str] | None = None,
    check_exit_code: bool | int | Iterable[int] | None = 0,
    delay_on_retry: bool = True,
    attempts: int = 1,
    run_as_root: bool = False,
    root_helper: str = '',
    shell: bool = False,
    loglevel: int = logging.DEBUG,
    log_errors: LogErrors = LogErrors.DEFAULT,
    binary: Literal[True] = ...,
    on_execute: Callable[..., Any] | None = None,
    on_completion: Callable[..., Any] | None = None,
    preexec_fn: Callable[..., Any] | None = None,
    prlimit: ProcessLimits | None = None,
    python_exec: str | None = None,
    timeout: int | None = None,
) -> tuple[bytes, bytes]: ...


@overload
def execute(
    *cmd: str,
    cwd: str | None = None,
    process_input: str | bytes | None = None,
    env_variables: dict[str, str] | None = None,
    check_exit_code: bool | int | Iterable[int] | None = 0,
    delay_on_retry: bool = True,
    attempts: int = 1,
    run_as_root: bool = False,
    root_helper: str = '',
    shell: bool = False,
    loglevel: int = logging.DEBUG,
    log_errors: LogErrors = LogErrors.DEFAULT,
    binary: bool = False,
    on_execute: Callable[..., Any] | None = None,
    on_completion: Callable[..., Any] | None = None,
    preexec_fn: Callable[..., Any] | None = None,
    prlimit: ProcessLimits | None = None,
    python_exec: str | None = None,
    timeout: int | None = None,
) -> tuple[str, str] | tuple[bytes, bytes]: ...


def execute(
    *cmd: str,
    cwd: str | None = None,
    process_input: str | bytes | None = None,
    env_variables: dict[str, str] | None = None,
    check_exit_code: bool | int | Iterable[int] | None = 0,
    delay_on_retry: bool = True,
    attempts: int = 1,
    run_as_root: bool = False,
    root_helper: str = '',
    shell: bool = False,
    loglevel: int = logging.DEBUG,
    log_errors: LogErrors = LogErrors.DEFAULT,
    binary: bool = False,
    on_execute: Callable[..., Any] | None = None,
    on_completion: Callable[..., Any] | None = None,
    preexec_fn: Callable[..., Any] | None = None,
    prlimit: ProcessLimits | None = None,
    python_exec: str | None = None,
    timeout: int | None = None,
) -> tuple[str, str] | tuple[bytes, bytes]:
    """Helper method to shell out and execute a command through subprocess.

    Allows optional retry.

    :param cmd:             Passed to subprocess.Popen.
    :param cwd:             Set the current working directory
    :param process_input:   Send to opened process.
    :param env_variables:   Environment variables and their values that
                            will be set for the process.
    :param check_exit_code: Single bool, int, or list of allowed exit
                            codes.  Defaults to [0].  Raise
                            :class:`ProcessExecutionError` unless
                            program exits with one of these code.
    :param delay_on_retry:  True | False. Defaults to True. If set to True,
                            wait a short amount of time before retrying.
    :param attempts:        How many times to retry cmd.
    :param run_as_root:     True | False. Defaults to False. If set to True,
                            the command is prefixed by the command specified
                            in the root_helper kwarg.
    :param root_helper:     command to prefix to commands called with
                            run_as_root=True
    :param shell:           whether or not there should be a shell used to
                            execute this command. Defaults to false.
    :param loglevel:        log level for execute commands.
    :param log_errors:      Should stdout and stderr be logged on error?
                            Possible values are
                            :py:attr:`~.LogErrors.DEFAULT`,
                            :py:attr:`~.LogErrors.FINAL`, or
                            :py:attr:`~.LogErrors.ALL`. Note that the
                            values :py:attr:`~.LogErrors.FINAL` and
                            :py:attr:`~.LogErrors.ALL`
                            are **only** relevant when multiple attempts of
                            command execution are requested using the
                            ``attempts`` parameter.
    :param binary:          On Python 3, return stdout and stderr as bytes if
                            binary is True, as Unicode otherwise.
    :param on_execute:      This function will be called upon process creation
                            with the object as a argument.  The Purpose of this
                            is to allow the caller of `processutils.execute` to
                            track process creation asynchronously.
    :param on_completion:   This function will be called upon process
                            completion with the object as a argument.  The
                            Purpose of this is to allow the caller of
                            `processutils.execute` to track process completion
                            asynchronously.
    :param preexec_fn:      This function will be called
                            in the child process just before the child
                            is executed. WARNING: On windows, we silently
                            drop this preexec_fn as it is not supported by
                            subprocess.Popen on windows (throws a
                            ValueError)
    :param prlimit:         Set resource limits on the child process. See
                            below for a detailed description.
    :param python_exec:     The python executable to use for enforcing
                            prlimits. If this is not set or is None, it will
                            default to use sys.executable.
    :param timeout:         Timeout (in seconds) to wait for the process
                            termination. If timeout is reached,
                            :class:`subprocess.TimeoutExpired` is raised.
    :returns:               (stdout, stderr) from process execution
    :raises:                :class:`UnknownArgumentError` on
                            receiving unknown arguments
    :raises:                :class:`ProcessExecutionError`
    :raises:                :class:`OSError`
    :raises:                :class:`subprocess.TimeoutExpired`

    The *prlimit* parameter can be used to set resource limits on the child
    process.  If this parameter is used, the child process will be spawned by a
    wrapper process which will set limits before spawning the command.

    .. versionchanged:: 3.17
       *process_input* can now be either bytes or string on python3.

    .. versionchanged:: 3.4
       Added *prlimit* optional parameter.

    .. versionchanged:: 1.5
       Added *cwd* optional parameter.

    .. versionchanged:: 1.9
       Added *binary* optional parameter. On Python 3, *stdout* and *stderr*
       are now returned as Unicode strings by default, or bytes if *binary* is
       true.

    .. versionchanged:: 2.1
       Added *on_execute* and *on_completion* optional parameters.

    .. versionchanged:: 2.3
       Added *preexec_fn* optional parameter.
    """

    if process_input is not None:
        process_input = encodeutils.to_utf8(process_input)

    if python_exec is None:
        python_exec = sys.executable

    ignore_exit_code = False

    if isinstance(check_exit_code, bool):
        ignore_exit_code = not check_exit_code
        check_exit_code = [0]
    elif isinstance(check_exit_code, int):
        check_exit_code = [check_exit_code]
    elif check_exit_code is None:
        check_exit_code = [0]

    if isinstance(log_errors, int):
        log_errors = LogErrors(log_errors)
    if not isinstance(log_errors, LogErrors):
        raise InvalidArgumentError(
            _('Got invalid arg log_errors: %r') % log_errors
        )

    if run_as_root and hasattr(os, 'geteuid') and os.geteuid() != 0:
        if not root_helper:
            raise NoRootWrapSpecified(
                message=_(
                    'Command requested root, but did not '
                    'specify a root helper.'
                )
            )
        if shell:
            # root helper has to be injected into the command string
            cmd = (' '.join((root_helper, cmd[0])),) + cmd[1:]
        else:
            # root helper has to be tokenized into argument list
            cmd = tuple(shlex.split(root_helper)) + cmd

    cmd = tuple(str(c) for c in cmd)

    if prlimit:
        args = [python_exec, '-m', 'oslo_concurrency.prlimit']
        args.extend(prlimit.prlimit_args())
        args.append('--')
        args.extend(cmd)
        cmd = tuple(args)

    sanitized_cmd = strutils.mask_password(' '.join(cmd))

    watch = timeutils.StopWatch()
    while attempts > 0:
        attempts -= 1
        watch.restart()

        try:
            LOG.log(loglevel, _('Running cmd (subprocess): %s'), sanitized_cmd)
            _PIPE = subprocess.PIPE

            on_preexec_fn = functools.partial(_subprocess_setup, preexec_fn)

            # we're okay with opening the user-provided command here
            obj = subprocess.Popen(  # noqa: S603
                cmd,
                stdin=_PIPE,
                stdout=_PIPE,
                stderr=_PIPE,
                close_fds=True,
                preexec_fn=on_preexec_fn,
                shell=shell,
                cwd=cwd,
                env=env_variables,
            )

            if on_execute:
                on_execute(obj)

            try:
                result = obj.communicate(process_input, timeout=timeout)

                if obj.stdin:
                    obj.stdin.close()

                _returncode = obj.returncode
                LOG.log(
                    loglevel,
                    'CMD "%s" returned: %s in %0.3fs',
                    sanitized_cmd,
                    _returncode,
                    watch.elapsed(),
                )
            except subprocess.TimeoutExpired:
                LOG.log(
                    loglevel,
                    'CMD "%s" reached timeout in %0.3fs',
                    sanitized_cmd,
                    watch.elapsed(),
                )
                raise
            finally:
                if on_completion:
                    on_completion(obj)

            if not ignore_exit_code and _returncode not in check_exit_code:
                _stdout, _stderr = result
                stdout = os.fsdecode(_stdout)
                stderr = os.fsdecode(_stderr)
                sanitized_stdout = strutils.mask_password(stdout)
                sanitized_stderr = strutils.mask_password(stderr)
                raise ProcessExecutionError(
                    exit_code=_returncode,
                    stdout=sanitized_stdout,
                    stderr=sanitized_stderr,
                    cmd=sanitized_cmd,
                )

            if not binary and result is not None:
                _stdout, _stderr = result
                # Decode from the locale using using the surrogateescape error
                # handler (decoding cannot fail)
                stdout = os.fsdecode(_stdout)
                stderr = os.fsdecode(_stderr)
                return (stdout, stderr)
            else:
                return result
        except (ProcessExecutionError, OSError) as err:
            # if we want to always log the errors or if this is
            # the final attempt that failed and we want to log that.
            if log_errors == LOG_ALL_ERRORS or (
                log_errors == LOG_FINAL_ERROR and not attempts
            ):
                if isinstance(err, ProcessExecutionError):
                    format = _(
                        '%(desc)r\ncommand: %(cmd)r\n'
                        'exit code: %(code)r\nstdout: %(stdout)r\n'
                        'stderr: %(stderr)r'
                    )
                    LOG.log(
                        loglevel,
                        format,
                        {
                            "desc": err.description,
                            "cmd": err.cmd,
                            "code": err.exit_code,
                            "stdout": err.stdout,
                            "stderr": err.stderr,
                        },
                    )
                else:
                    format = _(
                        'Got an OSError\ncommand: %(cmd)r\nerrno: %(errno)r'
                    )
                    LOG.log(
                        loglevel,
                        format,
                        {"cmd": sanitized_cmd, "errno": err.errno},
                    )

            if not attempts:
                LOG.log(loglevel, _('%r failed. Not Retrying.'), sanitized_cmd)
                raise
            else:
                LOG.log(loglevel, _('%r failed. Retrying.'), sanitized_cmd)
                if delay_on_retry:
                    # we're not using this for cryptographic purposes
                    time.sleep(random.randint(20, 200) / 100.0)  # noqa: S311
        finally:
            # NOTE(termie): this appears to be necessary to let the subprocess
            #               call clean something up in between calls, without
            #               it two execute calls in a row hangs the second one
            # NOTE(bnemec): termie's comment above is probably specific to the
            #               eventlet subprocess module, but since we still
            #               have to support that we're leaving the sleep.  It
            #               won't hurt anything in the stdlib case anyway.
            time.sleep(0)

    # This should never be reached, but mypy requires a return statement
    raise RuntimeError("Unexpected exit from retry loop")


def trycmd(
    *cmd: str,
    cwd: str | None = None,
    process_input: str | bytes | None = None,
    env_variables: dict[str, str] | None = None,
    check_exit_code: bool | int | Iterable[int] | None = 0,
    delay_on_retry: bool = True,
    attempts: int = 1,
    run_as_root: bool = False,
    root_helper: str = '',
    shell: bool = False,
    loglevel: int = logging.DEBUG,
    log_errors: LogErrors = LogErrors.DEFAULT,
    binary: bool = False,
    on_execute: Callable[..., Any] | None = None,
    on_completion: Callable[..., Any] | None = None,
    preexec_fn: Callable[..., Any] | None = None,
    prlimit: ProcessLimits | None = None,
    python_exec: str | None = None,
    timeout: int | None = None,
    discard_warnings: bool = False,
) -> tuple[str | bytes, str | bytes]:
    """A wrapper around execute() to more easily handle warnings and errors.

    Returns an (out, err) tuple of strings containing the output of
    the command's stdout and stderr.  If 'err' is not empty then the
    command can be considered to have failed.

    :param discard_warnings:  True | False. Defaults to False. If set to True,
                              then for succeeding commands, stderr is cleared
    :returns:                 (out, err) from process execution
    """
    try:
        out, err = execute(
            *cmd,
            cwd=cwd,
            process_input=process_input,
            env_variables=env_variables,
            check_exit_code=check_exit_code,
            delay_on_retry=delay_on_retry,
            attempts=attempts,
            run_as_root=run_as_root,
            root_helper=root_helper,
            shell=shell,
            loglevel=loglevel,
            log_errors=log_errors,
            binary=binary,
            on_execute=on_execute,
            on_completion=on_completion,
            preexec_fn=preexec_fn,
            prlimit=prlimit,
            python_exec=python_exec,
            timeout=timeout,
        )
        failed = False
    except ProcessExecutionError as exn:
        out, err = '', str(exn)
        failed = True

    if not failed and discard_warnings and err:
        # Handle commands that output to stderr but otherwise succeed
        err = ''

    return out, err


@overload
def ssh_execute(
    ssh: 'paramiko.client.SSHClient',
    cmd: str,
    process_input: bytes | None = None,
    addl_env: Any = None,
    check_exit_code: bool = True,
    binary: Literal[True] = True,
    timeout: int | None = None,
    sanitize_stdout: bool = True,
) -> tuple[bytes, bytes]: ...


@overload
def ssh_execute(
    ssh: 'paramiko.client.SSHClient',
    cmd: str,
    process_input: bytes | None = None,
    addl_env: Any = None,
    check_exit_code: bool = True,
    binary: Literal[False] = False,
    timeout: int | None = None,
    sanitize_stdout: bool = True,
) -> tuple[str, str]: ...


@overload
def ssh_execute(
    ssh: 'paramiko.client.SSHClient',
    cmd: str,
    process_input: bytes | None = None,
    addl_env: Any = None,
    check_exit_code: bool = True,
    binary: bool = False,
    timeout: int | None = None,
    sanitize_stdout: bool = True,
) -> tuple[str, str] | tuple[bytes, bytes]: ...


def ssh_execute(
    ssh: 'paramiko.client.SSHClient',
    cmd: str,
    process_input: bytes | None = None,
    addl_env: Any = None,
    check_exit_code: bool = True,
    binary: bool = False,
    timeout: int | None = None,
    sanitize_stdout: bool = True,
) -> tuple[str, str] | tuple[bytes, bytes]:
    """Run a command through SSH.

    :param ssh:             An SSH Connection object.
    :param cmd:             The command string to run.
    :param check_exit_code: If an exception should be raised for non-zero
                            exit.
    :param timeout:         Max time in secs to wait for command execution.
    :param sanitize_stdout: Defaults to True. If set to True, stdout is
                            sanitized i.e. any sensitive information like
                            password in command output will be masked.
    :returns:               (stdout, stderr) from command execution through
                            SSH.

    .. versionchanged:: 1.9
       Added *binary* optional parameter.
    """
    sanitized_cmd = strutils.mask_password(cmd)
    LOG.debug('Running cmd (SSH): %s', sanitized_cmd)
    if addl_env:
        raise InvalidArgumentError(_('Environment not supported over SSH'))

    if process_input:
        # This is (probably) fixable if we need it...
        raise InvalidArgumentError(_('process_input not supported over SSH'))

    stdin_stream, stdout_stream, stderr_stream = ssh.exec_command(
        cmd, timeout=timeout
    )
    channel = stdout_stream.channel

    # NOTE(justinsb): This seems suspicious...
    # ...other SSH clients have buffering issues with this approach
    _stdout = stdout_stream.read()
    _stderr = stderr_stream.read()

    stdin_stream.close()

    exit_status = channel.recv_exit_status()

    # Decode from the locale using using the surrogateescape error handler
    # (decoding cannot fail). Decode even if binary is True because
    # mask_password() requires Unicode on Python 3
    stdout = os.fsdecode(_stdout)
    stderr = os.fsdecode(_stderr)

    if sanitize_stdout:
        stdout = strutils.mask_password(stdout)

    stderr = strutils.mask_password(stderr)

    # exit_status == -1 if no exit code was returned
    if exit_status != -1:
        LOG.debug('Result was %d', exit_status)
        if check_exit_code and exit_status != 0:
            # In case of errors in command run, due to poor implementation of
            # command executable program, there might be chance that it leaks
            # sensitive information like password to stdout. In such cases
            # stdout needs to be sanitized even though sanitize_stdout=False.
            stdout = strutils.mask_password(stdout)
            raise ProcessExecutionError(
                exit_code=exit_status,
                stdout=stdout,
                stderr=stderr,
                cmd=sanitized_cmd,
            )

    if binary:
        # fsencode() is the reverse operation of fsdecode()
        return (os.fsencode(stdout), os.fsencode(stderr))
    else:
        return (stdout, stderr)


def get_worker_count() -> int:
    """Utility to get the default worker count.

    :returns: The number of CPUs if that can be determined, else a default
              worker count of 1 is returned.
    """
    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        return 1



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from oslo_config import fixture as config

from oslo_concurrency import lockutils


class LockFixture(fixtures.Fixture):
    """External locking fixture.

    This fixture is basically an alternative to the synchronized decorator with
    the external flag so that tearDowns and addCleanups will be included in
    the lock context for locking between tests. The fixture is recommended to
    be the first line in a test method, like so::

        def test_method(self):
            self.useFixture(LockFixture('lock_name'))
                ...

    or the first line in setUp if all the test methods in the class are
    required to be serialized. Something like::

        class TestCase(testtools.testcase):
            def setUp(self):
                self.useFixture(LockFixture('lock_name'))
                super(TestCase, self).setUp()
                    ...

    This is because addCleanups are put on a LIFO queue that gets run after the
    test method exits. (either by completing or raising an exception)
    """

    def __init__(
        self,
        name: str,
        lock_file_prefix: str | None = None,
    ):
        self.mgr = lockutils.lock(name, lock_file_prefix, True)

    def setUp(self) -> None:
        super().setUp()
        self.addCleanup(self.mgr.__exit__, None, None, None)
        self.lock = self.mgr.__enter__()


class ExternalLockFixture(fixtures.Fixture):
    """Configure lock_path so external locks can be used in unit tests.

    Creates a temporary directory to hold file locks and sets the oslo.config
    lock_path opt to use it.  This can be used to enable external locking
    on a per-test basis, rather than globally with the OSLO_LOCK_PATH
    environment variable.

    Example::

        def test_method(self):
            self.useFixture(ExternalLockFixture())
            something_that_needs_external_locks()

    Alternatively, the useFixture call could be placed in a test class's
    setUp method to provide this functionality to all tests in the class.

    .. versionadded:: 0.3
    """

    def setUp(self) -> None:
        super().setUp()
        temp_dir = self.useFixture(fixtures.TempDir())
        conf = self.useFixture(config.Config(lockutils.CONF)).config
        conf(lock_path=temp_dir.path, group='oslo_concurrency')



#    Copyright 2011 Justin Santa Barbara
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import errno
import fcntl
import multiprocessing
import os
import signal
import subprocess
import sys
import tempfile
import threading
import time
from unittest import mock

from oslotest import base as test_base

from oslo_concurrency.fixture import lockutils as fixtures
from oslo_concurrency import lockutils
from oslo_config import fixture as config


def lock_files(handles_dir, out_queue):
    with lockutils.lock('external', 'test-', external=True):
        # Open some files we can use for locking
        handles = []
        for n in range(50):
            path = os.path.join(handles_dir, (f'file-{n}'))
            handles.append(open(path, 'w'))

        # Loop over all the handles and try locking the file
        # without blocking, keep a count of how many files we
        # were able to lock and then unlock. If the lock fails
        # we get an IOError and bail out with bad exit code
        count = 0
        for handle in handles:
            try:
                fcntl.flock(handle, fcntl.LOCK_EX | fcntl.LOCK_NB)
                count += 1
                fcntl.flock(handle, fcntl.LOCK_UN)
            except OSError:
                os._exit(2)
            finally:
                handle.close()
        return out_queue.put(count)


class LockTestCase(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.config = self.useFixture(config.Config(lockutils.CONF)).config

    def test_synchronized_wrapped_function_metadata(self):
        @lockutils.synchronized('whatever', 'test-')
        def foo():
            """Bar."""
            pass

        self.assertEqual(
            'Bar.', foo.__doc__, "Wrapped function's docstring got lost"
        )
        self.assertEqual(
            'foo', foo.__name__, "Wrapped function's name got mangled"
        )

    def test_lock_internally_different_collections(self):
        s1 = lockutils.Semaphores()
        s2 = lockutils.Semaphores()
        trigger = threading.Event()
        who_ran: collections.deque[str] = collections.deque()

        def f(name, semaphores, pull_trigger):
            with lockutils.internal_lock('testing', semaphores=semaphores):
                if pull_trigger:
                    trigger.set()
                else:
                    trigger.wait()
                who_ran.append(name)

        threads = [
            threading.Thread(target=f, args=(1, s1, True)),
            threading.Thread(target=f, args=(2, s2, False)),
        ]
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        self.assertEqual([1, 2], sorted(who_ran))

    def test_lock_internally(self):
        """We can lock across multiple threads."""
        saved_sem_num = len(lockutils._semaphores)
        seen_threads = list()

        def f(_id):
            with lockutils.lock('testlock2', 'test-', external=False):
                for x in range(10):
                    seen_threads.append(_id)

        threads = []
        for i in range(10):
            thread = threading.Thread(target=f, args=(i,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        self.assertEqual(100, len(seen_threads))
        # Looking at the seen threads, split it into chunks of 10, and verify
        # that the last 9 match the first in each chunk.
        for i in range(10):
            for j in range(9):
                self.assertEqual(
                    seen_threads[i * 10], seen_threads[i * 10 + 1 + j]
                )

        self.assertEqual(
            saved_sem_num,
            len(lockutils._semaphores),
            "Semaphore leak detected",
        )

    def test_lock_internal_fair(self):
        """Check that we're actually fair."""

        def f(_id: int) -> None:
            with lockutils.lock(
                'testlock', 'test-', external=False, fair=True
            ):
                lock_holder.append(_id)

        lock_holder: list[int] = []
        threads = []
        # While holding the fair lock, spawn a bunch of threads that all try
        # to acquire the lock.  They will all block.  Then release the lock
        # and see what happens.
        with lockutils.lock('testlock', 'test-', external=False, fair=True):
            for i in range(10):
                thread = threading.Thread(target=f, args=(i,))
                threads.append(thread)
                thread.start()
                # Allow some time for the new thread to get queued onto the
                # list of pending writers before continuing.  This is gross
                # but there's no way around it without using knowledge of
                # fasteners internals.
                time.sleep(0.5)
        # Wait for all threads.
        for thread in threads:
            thread.join()

        self.assertEqual(10, len(lock_holder))
        # Check that the threads each got the lock in fair order.
        for i in range(10):
            self.assertEqual(i, lock_holder[i])

    def test_fair_lock_with_semaphore(self):
        def do_test():
            s = lockutils.Semaphores()
            with lockutils.lock('testlock', 'test-', semaphores=s, fair=True):
                pass

        self.assertRaises(NotImplementedError, do_test)

    def test_fair_lock_with_nonblocking(self):
        def do_test():
            with lockutils.lock(
                'testlock', 'test-', fair=True, blocking=False
            ):
                pass

        self.assertRaises(NotImplementedError, do_test)

    def test_nested_synchronized_external_works(self):
        """We can nest external syncs."""
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')
        sentinel = object()

        @lockutils.synchronized('testlock1', 'test-', external=True)
        def outer_lock():
            @lockutils.synchronized('testlock2', 'test-', external=True)
            def inner_lock():
                return sentinel

            return inner_lock()

        self.assertEqual(sentinel, outer_lock())

    def _do_test_lock_externally(self):
        """We can lock across multiple processes."""
        ctx = multiprocessing.get_context('fork')
        children = []
        for n in range(50):
            queue: multiprocessing.Queue[int] = ctx.Queue()
            proc = ctx.Process(
                target=lock_files, args=(tempfile.mkdtemp(), queue)
            )
            proc.start()
            children.append((proc, queue))
        for child, queue in children:
            child.join()
            count = queue.get(block=False)
            self.assertEqual(50, count)

    def test_lock_externally(self):
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

        self._do_test_lock_externally()

    def test_lock_externally_lock_dir_not_exist(self):
        lock_dir = tempfile.mkdtemp()
        os.rmdir(lock_dir)
        self.config(lock_path=lock_dir, group='oslo_concurrency')

        self._do_test_lock_externally()

    def test_lock_with_prefix(self):
        # TODO(efried): Embetter this test
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')
        foo = lockutils.lock_with_prefix('mypfix-')

        with foo('mylock', external=True):
            # We can't check much
            pass

    def test_synchronized_with_prefix(self):
        lock_name = 'mylock'
        lock_pfix = 'mypfix-'

        foo = lockutils.synchronized_with_prefix(lock_pfix)

        @foo(lock_name, external=True)
        def bar(dirpath, pfix, name):
            return True

        lock_dir = tempfile.mkdtemp()
        self.config(lock_path=lock_dir, group='oslo_concurrency')

        self.assertTrue(bar(lock_dir, lock_pfix, lock_name))

    def test_synchronized_without_prefix(self):
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

        @lockutils.synchronized('lock', external=True)
        def test_without_prefix():
            # We can't check much
            pass

        test_without_prefix()

    def test_synchronized_prefix_without_hypen(self):
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

        @lockutils.synchronized('lock', 'hypen', True)
        def test_without_hypen():
            # We can't check much
            pass

        test_without_hypen()

    def test_contextlock(self):
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

        # Note(flaper87): Lock is not external, which means
        # a semaphore will be yielded
        with lockutils.lock("test") as sem:
            self.assertIsInstance(sem, threading.Semaphore)

            # NOTE(flaper87): Lock is external so an InterProcessLock
            # will be yielded.
            with lockutils.lock("test2", external=True) as lock:
                self.assertTrue(lock.exists())

            with lockutils.lock("test1", external=True) as lock1:
                self.assertIsInstance(lock1, lockutils.InterProcessLock)

    def test_contextlock_unlocks(self):
        self.config(lock_path=tempfile.mkdtemp(), group='oslo_concurrency')

        with lockutils.lock("test") as sem:
            self.assertIsInstance(sem, threading.Semaphore)

            with lockutils.lock("test2", external=True) as lock:
                self.assertTrue(lock.exists())

            # NOTE(flaper87): Lock should be free
            with lockutils.lock("test2", external=True) as lock:
                self.assertTrue(lock.exists())

        # NOTE(flaper87): Lock should be free
        # but semaphore should already exist.
        with lockutils.lock("test") as sem2:
            self.assertEqual(sem, sem2)

    @mock.patch('logging.Logger.info')
    @mock.patch('os.remove')
    @mock.patch('oslo_concurrency.lockutils._get_lock_path')
    def test_remove_lock_external_file_exists(
        self, path_mock, remove_mock, log_mock
    ):
        lockutils.remove_external_lock_file(
            mock.sentinel.name, mock.sentinel.prefix, mock.sentinel.lock_path
        )

        path_mock.assert_called_once_with(
            mock.sentinel.name, mock.sentinel.prefix, mock.sentinel.lock_path
        )
        remove_mock.assert_called_once_with(path_mock.return_value)
        log_mock.assert_not_called()

    @mock.patch('logging.Logger.info')
    @mock.patch('os.remove')
    @mock.patch('oslo_concurrency.lockutils._get_lock_path')
    def test_remove_lock_external_file_exists_no_args(
        self, path_mock, remove_mock, log_mock
    ):
        lockutils.remove_external_lock_file(mock.sentinel.name)
        path_mock.assert_called_once_with(mock.sentinel.name, None, None)
        remove_mock.assert_called_once_with(path_mock.return_value)
        log_mock.assert_not_called()

    @mock.patch('logging.Logger.warning')
    @mock.patch('os.remove', side_effect=OSError(errno.ENOENT, None))
    @mock.patch('oslo_concurrency.lockutils._get_lock_path')
    def test_remove_lock_external_file_doesnt_exists(
        self, path_mock, remove_mock, log_mock
    ):
        lockutils.remove_external_lock_file(
            mock.sentinel.name, mock.sentinel.prefix, mock.sentinel.lock_path
        )
        path_mock.assert_called_once_with(
            mock.sentinel.name, mock.sentinel.prefix, mock.sentinel.lock_path
        )
        remove_mock.assert_called_once_with(path_mock.return_value)
        log_mock.assert_not_called()

    @mock.patch('logging.Logger.warning')
    @mock.patch('os.remove', side_effect=OSError(errno.EPERM, None))
    @mock.patch('oslo_concurrency.lockutils._get_lock_path')
    def test_remove_lock_external_file_permission_error(
        self, path_mock, remove_mock, log_mock
    ):
        lockutils.remove_external_lock_file(
            mock.sentinel.name, mock.sentinel.prefix, mock.sentinel.lock_path
        )
        path_mock.assert_called_once_with(
            mock.sentinel.name, mock.sentinel.prefix, mock.sentinel.lock_path
        )
        remove_mock.assert_called_once_with(path_mock.return_value)
        log_mock.assert_called()

    def test_no_slash_in_b64(self):
        # base64(sha1(foobar)) has a slash in it
        with lockutils.lock("foobar"):
            pass


class FileBasedLockingTestCase(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.lock_dir = tempfile.mkdtemp()

    def test_lock_file_exists(self):
        lock_file = os.path.join(self.lock_dir, 'lock-file')

        @lockutils.synchronized(
            'lock-file', external=True, lock_path=self.lock_dir
        )
        def foo():
            self.assertTrue(os.path.exists(lock_file))

        foo()

    def test_interprocess_lock(self):
        lock_file = os.path.join(self.lock_dir, 'processlock')

        pid = os.fork()
        if pid:
            # Make sure the child grabs the lock first
            start = time.time()
            while not os.path.exists(lock_file):
                if time.time() - start > 5:
                    self.fail('Timed out waiting for child to grab lock')
                time.sleep(0)
            lock1 = lockutils.InterProcessLock('foo')
            lock1.lockfile = open(lock_file, 'w')
            # NOTE(bnemec): There is a brief window between when the lock file
            # is created and when it actually becomes locked.  If we happen to
            # context switch in that window we may succeed in locking the
            # file.  Keep retrying until we either get the expected exception
            # or timeout waiting.
            while time.time() - start < 5:
                try:
                    lock1.trylock()
                    lock1.unlock()
                    time.sleep(0)
                except OSError:
                    # This is what we expect to happen
                    break
            else:
                self.fail('Never caught expected lock exception')
            # We don't need to wait for the full sleep in the child here
            os.kill(pid, signal.SIGKILL)
        else:
            try:
                lock2 = lockutils.InterProcessLock('foo')
                lock2.lockfile = open(lock_file, 'w')
                have_lock = False
                while not have_lock:
                    try:
                        lock2.trylock()
                        have_lock = True
                    except OSError:
                        pass
            finally:
                # NOTE(bnemec): This is racy, but I don't want to add any
                # synchronization primitives that might mask a problem
                # with the one we're trying to test here.
                time.sleep(0.5)
                os._exit(0)

    def test_interprocess_nonblocking_external_lock(self):
        """Check that we're not actually blocking between processes."""

        ctx = multiprocessing.get_context('fork')
        nb_calls = ctx.Value('i', 0)

        @lockutils.synchronized(
            'foo', blocking=False, external=True, lock_path=self.lock_dir
        )
        def foo(param):
            """Simulate a long-running operation in a process."""
            param.value += 1
            time.sleep(0.5)

        def other(param):
            foo(param)

        process = ctx.Process(target=other, args=(nb_calls,))
        process.start()
        # Make sure the other process grabs the lock
        start = time.time()
        while not os.path.exists(os.path.join(self.lock_dir, 'foo')):
            if time.time() - start > 5:
                self.fail('Timed out waiting for process to grab lock')
            time.sleep(0)
        process1 = ctx.Process(target=other, args=(nb_calls,))
        process1.start()
        process1.join()
        process.join()
        self.assertEqual(1, nb_calls.value)

    def test_interthread_external_lock(self):
        call_list = []

        @lockutils.synchronized('foo', external=True, lock_path=self.lock_dir)
        def foo(param):
            """Simulate a long-running threaded operation."""
            call_list.append(param)
            # NOTE(bnemec): This is racy, but I don't want to add any
            # synchronization primitives that might mask a problem
            # with the one we're trying to test here.
            time.sleep(0.5)
            call_list.append(param)

        def other(param):
            foo(param)

        thread = threading.Thread(target=other, args=('other',))
        thread.start()
        # Make sure the other thread grabs the lock
        # NOTE(bnemec): File locks do not actually work between threads, so
        # this test is verifying that the local semaphore is still enforcing
        # external locks in that case.  This means this test does not have
        # the same race problem as the process test above because when the
        # file is created the semaphore has already been grabbed.
        start = time.time()
        while not os.path.exists(os.path.join(self.lock_dir, 'foo')):
            if time.time() - start > 5:
                self.fail('Timed out waiting for thread to grab lock')
            time.sleep(0)
        thread1 = threading.Thread(target=other, args=('main',))
        thread1.start()
        thread1.join()
        thread.join()
        self.assertEqual(['other', 'other', 'main', 'main'], call_list)

    def test_interthread_nonblocking_external_lock(self):
        call_list = []

        @lockutils.synchronized(
            'foo', external=True, blocking=False, lock_path=self.lock_dir
        )
        def foo(param):
            """Simulate a long-running threaded operation."""
            call_list.append(param)
            time.sleep(0.5)
            call_list.append(param)

        def other(param):
            foo(param)

        thread = threading.Thread(target=other, args=('other',))
        thread.start()
        # Make sure the other thread grabs the lock
        start = time.time()
        while not os.path.exists(os.path.join(self.lock_dir, 'foo')):
            if time.time() - start > 5:
                self.fail('Timed out waiting for thread to grab lock')
            time.sleep(0)
        thread1 = threading.Thread(target=other, args=('main',))
        thread1.start()
        thread1.join()
        thread.join()
        self.assertEqual(['other', 'other'], call_list)

    def test_interthread_nonblocking_internal_lock(self):
        call_list = []

        @lockutils.synchronized('foo', blocking=False, lock_path=self.lock_dir)
        def foo(param):
            # Simulate a long-running threaded operation.
            call_list.append(param)
            time.sleep(0.5)
            call_list.append(param)

        def other(param):
            foo(param)

        thread = threading.Thread(target=other, args=('other',))
        thread.start()
        # Make sure the other thread grabs the lock
        start = time.time()
        while not call_list:
            if time.time() - start > 5:
                self.fail('Timed out waiting for thread to grab lock')
            time.sleep(0)
        thread1 = threading.Thread(target=other, args=('main',))
        thread1.start()
        thread1.join()
        thread.join()
        self.assertEqual(['other', 'other'], call_list)

    def test_non_destructive(self):
        lock_file = os.path.join(self.lock_dir, 'not-destroyed')
        with open(lock_file, 'w') as f:
            f.write('test')
        with lockutils.lock(
            'not-destroyed', external=True, lock_path=self.lock_dir
        ):
            with open(lock_file) as f:
                self.assertEqual('test', f.read())


class LockutilsModuleTestCase(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.old_env = os.environ.get('OSLO_LOCK_PATH')
        if self.old_env is not None:
            del os.environ['OSLO_LOCK_PATH']

    def tearDown(self):
        if self.old_env is not None:
            os.environ['OSLO_LOCK_PATH'] = self.old_env
        super().tearDown()

    def test_main(self):
        script = '\n'.join(
            [
                'import os',
                'lock_path = os.environ.get("OSLO_LOCK_PATH")',
                'assert lock_path is not None',
                'assert os.path.isdir(lock_path)',
            ]
        )
        argv = ['', sys.executable, '-c', script]
        retval = lockutils._lock_wrapper(argv)
        self.assertEqual(0, retval, "Bad OSLO_LOCK_PATH has been set")

    def test_return_value_maintained(self):
        script = '\n'.join(
            [
                'import sys',
                'sys.exit(1)',
            ]
        )
        argv = ['', sys.executable, '-c', script]
        retval = lockutils._lock_wrapper(argv)
        self.assertEqual(1, retval)

    def test_direct_call_explodes(self):
        cmd = [sys.executable, '-m', 'oslo_concurrency.lockutils']
        with open(os.devnull, 'w') as devnull:
            retval = subprocess.call(cmd, stderr=devnull)
            self.assertEqual(1, retval)


class TestLockFixture(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.config = self.useFixture(config.Config(lockutils.CONF)).config
        self.tempdir = tempfile.mkdtemp()

    def _check_in_lock(self):
        self.assertTrue(self.lock.exists())

    def tearDown(self):
        self._check_in_lock()
        super().tearDown()

    def test_lock_fixture(self):
        # Setup lock fixture to test that teardown is inside the lock
        self.config(lock_path=self.tempdir, group='oslo_concurrency')
        fixture = fixtures.LockFixture('test-lock')
        self.useFixture(fixture)
        self.lock = fixture.lock


class TestGetLockPath(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.conf = self.useFixture(config.Config(lockutils.CONF)).conf

    def test_get_default(self):
        lockutils.set_defaults(lock_path='/the/path')
        self.assertEqual('/the/path', lockutils.get_lock_path(self.conf))

    def test_get_override(self):
        lockutils._register_opts(self.conf)
        self.conf.set_override(
            'lock_path', '/alternate/path', group='oslo_concurrency'
        )
        self.assertEqual('/alternate/path', lockutils.get_lock_path(self.conf))
